#ifndef INTERRUPT_H
#define INTERRUPT_H

//
// Class: Interrupt
//
// The Interrupt class is very simple. It provides the Interrupt type
// enum.
// 
// Notes: I had planned for Interrupt to include a pointer to the Process 
// which "created" or caused the Interrupt. I did not because it would have 
// caused a circular reference. I believe that all this would go away if 
// CPU and I/O devices properly raised interrupts, rather than the process
// doing do directly.
//

class Interrupt { 
public:
	enum Type { CpuComplete=0, TimeSliceComplete=1, IoRequest=2, IoComplete=3 } ; 

	Interrupt( Type t) { 
		iType = t ; 
	}

	Type getType() { return iType ; }

	// String output for stats and debugging
	char *getTypeString() { 
		char *s ; 
		switch( iType) { 
		case CpuComplete: s = "CPU complete" ; break ; 
		case TimeSliceComplete: s = "Time Slice"; break ; 
		case IoRequest: s = "IO request"; break ; 
		case IoComplete: s = "IO complete"; break ; 
		default: s = "Bad interrupt type!"; break ; 
		}
		return s ; 
	}

private: 
	Type iType ;		// that's it
} ;

#endif
