#include <iostream.h>
#include <fstream>
#include <queue>
#include <list>
#include "process.h"
#include "cpu.h"
#include "billos.h"

using namespace std ; 

void try2( fstream &, fstream &) ; 
Process *readProcessFile( fstream &, fstream &) ; 
static bool simulationDone( Cpu, queue<Process *>, list<Process *>) ; 
static void debugInterrupt( Interrupt *, Process *, int, fstream &) ; 


static bool debugInts = true ; 

int main() { 
	char		fileName[ 100] ; 
//	char		ch ; 
//	char		curLine[ 100] ; 
	fstream		dataFile ; 
	fstream		outFile ; 
	Process		*p ; 
	queue<Process *>	readyQ ; 
	list<Process *>		terminatedList ; 
	list<Process *>::iterator	terminatedIter ; 
	int			systemClock ; 
	Cpu			theCpu ; 
	Interrupt	*cpuInterrupt ; 
	Interrupt	*ioInterrupt ; 
	int			timeSlice = 10 ; 
	Process		*holdProcessForIo ; 
	list<Process *>				ioList ; 
	list<Process *>::iterator	ioIter ; 
	Process		*removeIoProcess ; 

	// Open the data set file
	cout << "Enter the name of the data set file: " ; 
	cin.getline( fileName, 100) ; 
	dataFile.open( fileName, ios::in) ; 
	if( !dataFile) { 
		cout << "Error: Bad data set file" ; 
		exit( 1) ; 
	}

	// Open the output file
	outFile.open( "output.txt", ios::out) ; 
	if( !outFile) { 
		cout << "Error: Can't open output file output.txt" << endl ; 
		exit( 1) ; 
	}
	outFile << "DATA SET: " << fileName << endl ; 

	try2( dataFile, outFile) ; 
	exit( 0) ; 

	while( ! dataFile.eof()) { 
		p = readProcessFile( dataFile, outFile) ; 
		if( p != 0) { 
			readyQ.push( p) ; 
		}
	}

	outFile << "\n\nDEBUG:" << endl ; 

//	cout << "In beginning, readyQ size=" << readyQ.size() << endl ; 

	// initialization before the main simulation loop
	p = readyQ.front() ; 
	readyQ.pop() ; 
	theCpu.setRunningProcess( p) ; 
//	Process *nullProcess = new Process() ; 
//	ioList.push_back( nullProcess) ; 
	systemClock = 0 ; 

	// main simulation loop
	while( !simulationDone( theCpu, readyQ, ioList)) {
		systemClock++ ; 
		
		holdProcessForIo = 0 ; 

		// Run the CPU process for one time unit
		cpuInterrupt = theCpu.doBurst() ; 

		if( cpuInterrupt == 0) { 
			// Check if the time slice is up for this process
			if( theCpu.getRunTotal() == timeSlice) { 
				cpuInterrupt = new Interrupt( Interrupt::TimeSliceComplete) ; 
			}
		}

		// Switch handles all interrupts that can be generated when 
		// a process is running on the CPU
		if( cpuInterrupt != 0) {
			p = theCpu.getRunningProcess() ; 
			theCpu.removeRunningProcess() ; 

			debugInterrupt( cpuInterrupt, p, systemClock, outFile) ; 

			switch( cpuInterrupt->getType()) { 
			case Interrupt::CpuComplete:
				p->setTerminatedTime( systemClock) ; 
				terminatedList.push_back( p) ; 
//				cout << "Process " << p->getId() ; 
//				cout << " completed at time=" << systemClock ; 
//				cout << " with a wait=" << p->calcWaitTime( systemClock) << endl ; 
				break ; 

			case Interrupt::IoRequest:
				holdProcessForIo = p ; 
				break ; 

			case Interrupt::TimeSliceComplete:
				readyQ.push( p) ; 
				break ; 
				
			default: 
				cout << "Error: Bad CPU interrupt, type=" << cpuInterrupt->getType() << endl ; 
				break ; 
			}
		}

		// Do I/O for all processes in the I/O list
		if( !ioList.empty()) { 
			ioIter = ioList.begin() ; 
			while( ioIter != ioList.end()) { 
				removeIoProcess = 0 ; 
				p = *ioIter ; 
				ioInterrupt = p->doIoBurst() ; 

				if( ioInterrupt != 0) {
					debugInterrupt( ioInterrupt, p, systemClock, outFile) ; 

					switch( ioInterrupt->getType()) { 
					case Interrupt::IoComplete:
						removeIoProcess = p ; 
//						ioList.erase( ioIter) ; 
						readyQ.push( p) ; 
						break ; 
					default: 
						cout << "Error: Bad I/O interrupt, type=" << ioInterrupt->getType() << endl ; 
						break ; 
					}
				}
				if( ioIter != ioList.end()) { 
					ioIter++ ; 
				}
				if( removeIoProcess != 0) { 
					ioList.remove( removeIoProcess) ; 
					removeIoProcess = 0 ; 
				}
			}
		}

		// This is a cleeb. If a process has been held for inclusion on 
		// the I/O list from above, add it now.
		if( holdProcessForIo != 0) { 
			ioList.push_back( holdProcessForIo) ; 
//			cout << "At time=" << systemClock ; 
//			cout << " add to I/O list, process=" << holdProcessForIo->getId() << endl ; 
		}

		// If, after all is said and done, the CPU is not busy, then 
		// pop a process off the ReadyQ
		if( !theCpu.isBusy()) { 
			if( !readyQ.empty()) { 
				p = readyQ.front() ; 
				readyQ.pop() ; 
				theCpu.setRunningProcess( p) ; 
			}
		}
	}

	// Print out the stats of all the completed/terminated processes
	outFile << "\n\nSTATS:" << endl ; 
	terminatedIter = terminatedList.begin() ; 
	while( terminatedIter != terminatedList.end()) { 
		p = *terminatedIter ; 

		outFile << "Process " << p->getId() ; 
		outFile << " completed at time " << p->getTerminatedTime() ; 
		outFile << " with a wait of " << p->calcWaitTime() << endl ; 

		terminatedIter++ ; 
	}
	outFile.close() ; 

	cout << "Simulation complete at system time " << systemClock << endl ; 
	cout << "Hello world." << endl ; 
	return 0 ; 
}

Process *readProcessFile( fstream &inFile, fstream &outFile) {
	int	id, cpuRequired, ioTime1, ioLength1, ioTime2, ioLength2 ; 

	if( inFile.eof()) { return 0 ; }

	inFile >> id ; 
	if( inFile.eof()) { return 0 ; }
	inFile >> cpuRequired ; 
	inFile >> ioTime1 ; 
	inFile >> ioLength1 ; 
	inFile >> ioTime2 ; 
	inFile >> ioLength2 ; 

	outFile << id << "\t" << cpuRequired << "\t" ;
	outFile << ioTime1 << "\t" << ioLength1 << "\t" ; 
	outFile << ioTime2 << "\t" << ioLength2 << endl ; 

	// Create a new process with this information
	Process *p = new Process( id, cpuRequired) ; 
	p->setIoRequest1( ioTime1, ioLength1) ; 
	p->setIoRequest2( ioTime2, ioLength2) ; 
	return p ; 
}

static bool simulationDone( Cpu theCpu, queue<Process *> readyQ, 
						   list<Process *> ioList) {
	if( !theCpu.isBusy()  &&
		readyQ.empty()  &&  ioList.empty()) { 
//		ioList.size() <= 1) { 
		return true ; 
	}
	return false ; 
}

static void debugInterrupt( Interrupt *intr, Process *p, int clock, 
						   fstream &outFile) {
	if( debugInts) { 
		outFile << clock << ": " ; 
		outFile << "p(" << p->getId() << ") " ; 
		outFile << intr->getTypeString() << " intr" ; 
		outFile << " - cpu(" << p->getId() << ")= " << p->getCpuCompleted() << endl ; 
	}
}

void try2( fstream &dataFile, fstream &outFile) { 
	BillOs		theOS ; 
	char		c ; 

	theOS.setTimeSlice( 10) ; 

	theOS.addProcessFile( dataFile, outFile) ; 
	
	theOS.init() ; 
	cout << "init" << endl ; 
	cout << "Hit enter..." ; 
	cin.get( c) ; 

	theOS.run() ; 
	cout << "run" << endl ; 
	cout << "Hit enter..." ; 
	cin.get( c) ; 

	theOS.outputStats( outFile) ; 
	outFile.close() ; 
}
