#include <iostream.h>
#include "Process.h"

//using namespace std ; 

// Method: Constructor
// With no args, a "null" Process is created.  It has no ID or CPU requirements.
Process::Process() { 
	pId = 0 ; 
	pState = pNew ; 
	pCpuRequired = 0 ; 
	pCpuCompleted = 0 ; 
	pTerminatedTime = 0 ; 

	pIoRequestTime1 = 0 ; 
	pIoRequestLength1 = 0 ; 
	pIoCompleted1 = 0 ; 
	pIoRequestTime2 = 0 ;
	pIoRequestLength2 = 0 ; 
	pIoCompleted2 = 0 ; 
}

// Method: Constructor
// With args, the construcotr creates a "real" Process with ID and CPU requirements.
Process::Process( int id, int cpuRequired) { 
	pId = id ; 
	pState = pNew ; 

	pCpuRequired = cpuRequired ; 
	pCpuCompleted = 0 ; 
	pTerminatedTime = 0 ; 

	pIoRequestTime1 = 0 ; 
	pIoRequestLength1 = 0 ; 
	pIoCompleted1 = 0 ; 
	pIoRequestTime2 = 0 ;
	pIoRequestLength2 = 0 ; 
	pIoCompleted2 = 0 ; 
}

// Method: setIoRequest1, setIoRequest2
// Set the time and length of the Process' 1st (or 2nd) I/O request
// Notes: Should error check these values against CPU required
void Process::setIoRequest1( int time, int length) {
	pIoRequestTime1 = time ; 
	pIoRequestLength1 = length ; 
	pIoCompleted1 = 0 ; 
}

void Process::setIoRequest2( int time, int length) {
	pIoRequestTime2 = time ; 
	pIoRequestLength2 = length ; 
	pIoCompleted2 = 0 ; 
}

// Method: doCpuBurst
// Executes the Process for 1 time unit of CPU, doing necessary Process accounting.
// The return value is 0 if operation is normal and continuing.
// If the Process needs attention, the appropriate Interrupt will be returned.
Interrupt *Process::doCpuBurst() { 
	Interrupt *i = 0 ; 

	pCpuCompleted++ ; 
	if( pCpuCompleted == pCpuRequired) { 
		// Completed Process' total CPU requirement
		i = new Interrupt( Interrupt::CpuComplete) ; 
	}
	else if( pCpuCompleted == pIoRequestTime1 ||
		pCpuCompleted == pIoRequestTime2) { 
		// Process needs I/O now
		i = new Interrupt( Interrupt::IoRequest) ; 
	}

	return i ; 
}

// Method: doIoBurst
// Performs 1 time unit of I/O for the Process, doing necessary accounting.
// The return value is 0 if operation is normal and continuing.
// If the Process needs attention, the appropriate Interrupt will be returned.
// Notes: This is very similar to doCpuBurst()
Interrupt *Process::doIoBurst() { 
	Interrupt *i = 0 ;	// default return is 0, normal operation

	if( pCpuCompleted == pIoRequestTime1) { 
		pIoCompleted1++ ; 
		if( pIoCompleted1 == pIoRequestLength1) {	// First I/O is completed
			i = new Interrupt( Interrupt::IoComplete) ; 
		}
	}
	else if( pCpuCompleted == pIoRequestTime2) { 
		pIoCompleted2++ ; 
		if( pIoCompleted2 == pIoRequestLength2) {	// Second I/O is completed
			i = new Interrupt( Interrupt::IoComplete) ; 
		}
	}
	else { 
		cout << "Error: Doing IO when Cpu time does not match!" << endl ; 
	}
	return i ; 
}

// Method: calcWaitTime
// Returns the total time by the Process not on CPU or doing I/O.
// This method only makes sense after the Process is terminated.
int Process::calcWaitTime() { 
	int wait ; 
	wait = pTerminatedTime - pCpuRequired ; 
	if( pIoRequestLength1 > 0) { 
		wait = wait - pIoRequestLength1 ; 
	}
	if( pIoRequestLength2 > 0) { 
		wait = wait - pIoRequestLength2 ; 
	}
	return wait ; 
}
