#include <iostream>
#include <fstream>
#include <vector>
#include <assert.h>

using namespace std ; 

#include "vmsimulator.h"


const int		vmAddressBits = 8 ; 

vector<int>		getDataSet( char *fileName, fstream &) ; 
void			simulateDataSet( int numFrames, 
								vector<int> referenceList, fstream &) ; 

//
//
int main() { 
	const int		maxNameSize = 100 ;
	char			fileName[ maxNameSize + 1] ; 
	vector<int>		referenceList ; 
	fstream			outFile ; 
	int				numFrames = 0 ; 

	// Open the output file for reporting
	outFile.open( "output.txt", ios::out) ; 
	if( !outFile) { 
		cout << "Error: Can't open output file output.txt" << endl ; 
		exit( 1) ; 
	}

	// Get the number of frames for this siumulation
	cout << "Enter the number of frames for this simulation: " ; 
	cin >> numFrames ; 
	assert( numFrames > 0) ; 
	outFile << "Number of frames: " << numFrames << endl ; 

	// Get the input file and read the reference string... our data set
	cout << "Enter the name of the data set file: " ; 
	cin >> fileName ; 
//	cin.getline( fileName, maxNameSize) ; 
	referenceList = getDataSet( fileName, outFile) ; 
	assert( referenceList.size() > 0) ; 
	outFile << endl << endl ; 

	// For each of the algorithms we're testing, create a simulator 
	// and run it on the data set
	simulateDataSet( numFrames, referenceList, outFile) ; 

	outFile.close() ; 
	cout << endl << "Simulation results available in output.txt" << endl ; 
	return 0 ; 
}

///////
// getDataSet()
//
vector<int> getDataSet( char *fileName, fstream &outFile) {
	fstream		dataFile ; 
	int			ref ; 
	vector<int>	referenceList ; 	

	// Open the data set file
	outFile << "DATA SET: " << fileName << endl ; 
	dataFile.open( fileName, ios::in) ; 
	if( !dataFile) { 
		cout << "Error: Bad data set file=" << fileName << endl ; 
		return referenceList ; 
	}

	outFile << "\t" ; 
	while( !dataFile.eof()) { 
		ref = -1 ; 
		dataFile >> ref ; 
		if( ref >= 0) {
			outFile << ref << " " ; 
			referenceList.push_back( ref) ; 
		}
	}
	dataFile.close() ; 
	outFile << endl ; 

	cout << "Number of refs=" << referenceList.size() << endl ; 
	return referenceList; 
}

///////
//
//
void simulateDataSet( int numFrames, vector<int> referenceList, fstream &outFile) { 
	VMSimulator		*vms ; 
	bool			simulationOK = true ; 

	// Do the FIFO simulation
	outFile << "Simulation Algorithm: FIFO" << endl ; 
	vms = new VMSimulator( vmAddressBits, numFrames, VMSimulator::FIFO) ; 
	simulationOK = vms->simulatePageReferences( referenceList, outFile) ; 
	if( simulationOK) { 
		vms->report( outFile) ; 
	}
	else { 
		outFile << "FIFO Simulation ended with an error" << endl ; 
	}

	// Do the LRU simulation
	outFile << endl << "Simulation Algorithm: LRU" << endl ; 
	vms = new VMSimulator( vmAddressBits, numFrames, VMSimulator::LRU) ; 
	simulationOK = vms->simulatePageReferences( referenceList, outFile) ; 
	if( simulationOK) { 
		vms->report( outFile) ; 
	}
	else { 
		outFile << "LRU Simulation ended with an error" << endl ; 
	}

	// Do the Optimal simulation
	outFile << endl << "Simulation algorithm: Optimal" << endl ; 
	vms = new VMSimulator( vmAddressBits, numFrames, VMSimulator::Optimal) ; 
	simulationOK = vms->simulatePageReferences( referenceList, outFile) ; 
	if( simulationOK) { 
		vms->report( outFile) ; 
	}
	else { 
		outFile << "Optimal Simulation ended with an error" << endl ; 
	}
}
