#include <iostream.h>
#include <fstream>
#include <assert.h>
#include <math.h>

using namespace std ; 

#include "pagetable.h"

//
PageTable::PageTable( int addressBits) { 
	pAddressBits = addressBits ; 

	pTableSize = (int) pow( 2.0, (double) pAddressBits) ; 
	cout << "pt: bits=" << pAddressBits << ", size=" << pTableSize << endl ; 

	pTable = new PageTableEntry[ pTableSize] ; 
	assert( pTable != 0) ; 

	for( int p = 0; p < pTableSize; p++) { 
		pTable[ p].frame = -1 ; 
		pTable[ p].timeOfUse = -1 ; 
		pTable[ p].valid = false ; 
		pTable[ p].dirty = false ; 
	}
}

//
bool PageTable::isPageInRange( int page) { 
	if( page >= 0  &&  page < pTableSize) { return true ; }
	else { return false ; }
}

//
bool PageTable::isPageValid( int page) { 
	bool	valid = false ; 

	if( isPageInRange( page)) { 
		valid = pTable[ page].valid ; 
	}
	return valid ; 
}

///////
//
int PageTable::getFrame( int page) { 
	int		frame = -1 ; 

	if( isPageInRange( page)  &&  pTable[page].valid == true) { 
		frame = pTable[ page].frame ; 
	}

	return frame ; 
}

///////
//
void PageTable::setFrame( int page, int frame) { 
	if( isPageInRange( page)) { 
		pTable[ page].valid = true ; 
		pTable[ page].frame = frame ; 
	}
}

// NOTES: Bill, why don't you be a man and just free the frame properly,
// and then realloc it to the new page. You weenie.
void PageTable::unsetFrame( int page) { 

	if( ! isPageInRange( page)) { return ; }

	pTable[ page].valid = false ; 
	pTable[ page].frame = -1 ; 
}

///////
//
int PageTable::getTimeOfUse( int page) { 
	int		time = -1 ; 

	if( isPageInRange( page)  &&  pTable[page].valid == true) { 
		time = pTable[ page].timeOfUse ; 
	}

	return time ; 
}

///////
//
void PageTable::setTimeOfUse( int page, int time) { 
	if( isPageInRange( page)) { 
		pTable[ page].timeOfUse = time ; 
	}
}

///////
//
void PageTable::debug( fstream &outFile) { 
	int		frame ; 
	for( int p = 0; p < pTableSize; p++) { 
		frame = pTable[ p].frame ; 
		if( frame >= 0) { 
			outFile << "(" << p << "," << frame << ") " ; 
		}
	}
}
