#ifndef VMSimulator_H
#define VMSimulator_H

#include <vector>
#include <queue>
#include "pagetable.h"
#include "framelist.h"

using namespace std ;

// Class: VMSimulator
// This class simulates virtual memory allocation
//
// NOTES: The "page replacement algorithm" should be an external function, 
// ala sort or the other STL algorithms.
//
class VMSimulator {
public:

	enum PageReplacementAlgorithm {Optimal, FIFO, LRU} ; 

	// Constructor - Three requirements before any simulation:
	// 1) You must define the number of bits in the VM address as this 
	// defines the size of virtual memory.
	// 2) You have to define the number of frames in the simulation.
	// 3) You have to select the page replacement algorithm
	VMSimulator( int addressBits, int numFrames, PageReplacementAlgorithm algo) ; 

	int		getAddressBits() { return vAddressBits ; }  
	int		getVirtualMemorySize() { return vMemorySize ; }
	int		getNumFrames() { return vNumFrames ; }

	// Reference a page in virtual memory. This is where the simulation happens.
	// This has to return an error for bad page addr!
	bool	simulatePageReference( int pageAddr, fstream &outFile) ;
	bool	simulatePageReferences( vector<int> pageAddrs, fstream &outFile) ; 

	void	report( fstream &outFile) ; 
	int		getNumPageFaults() { return vNumPageFaults ; }

private:
	int			vAddressBits ; 
	int			vNumFrames ; 
	PageReplacementAlgorithm	vAlgorithm ; 
	int			vMemorySize ; 

	PageTable	*vPageTable ; 
	FrameList	*vFrameList ; 
	int			vNumPageFaults ; 

	// Fifo fields
	queue<int>	vFifoQ ; 
//	list<int>	vLruList ; 

	// Private (local) methods
	int		findFifoVictim( int) ; 
	int		findLruVictim( int) ; 
	int		findOptimalVictim( int, vector<int>, int) ; 

} ; 

#endif
