#include <Windows.h>
#include <time.h>
#include <iostream>
#include <fstream>
#include <string>
#include <queue>

using namespace std ; 


// Thread function stubs
void			doSimpleton() ; 
DWORD WINAPI	simpleton( LPVOID) ; 
void			doSmarty() ; 
DWORD WINAPI	smarty( LPVOID) ; 
void			doBoundedBuffer() ; 
DWORD WINAPI	consumer( LPVOID) ; 
DWORD WINAPI	producer( LPVOID) ; 

// Utility function stubs
bool			getUserOptions() ; 
int				queryUserForInt( char *) ; 
string			queryUserForString( char *) ; 
void			gracefulExit( int) ; 

// Local synch objects
HANDLE			mutex = NULL ; 
HANDLE			full = NULL ; 
HANDLE			empty = NULL ; 

const int		MAX_THREADS = 12 ;	// limit on # threads allowed
const int		MAX_BUFFERS = 100 ; // buffer limit
const int		MAX_PRODUCER_TIME = 50 ; 
const int		MAX_CONSUMER_WAIT = 1000 ; 

int				threadCount = 0 ; 
int				numThreads = 0 ; 
string			outFileName ; 
fstream			outFile ; 
int				numConsumers = 0 ; 
int				numProducers = 0 ; 
int				itemsPerProducer = 0 ; 
int				maxProducerTime = 0 ; 
int				maxConsumerWait = 0 ; 
int				numBuffers = 0 ; 
queue<int>		bufferQ ;		// Bounded buffer Q

///////
// prog03 main()
// Prof. Bill Krieger
//
// CSC 420 Program #3 is a testbed for creating multithreaded 
// functions using the Win32 API.
// NOTE: This program was morphed from Prof Renk's "Dining Philosopher"
// code from his CSC 420 class.
//
int main() { 
	int			testCase ; 

	// Get the test case, call the appropriate test function
	testCase = queryUserForInt( "Select test (1=simpleton,2=smarty,3=bounded buffer) to run") ; 

	switch( testCase) { 
	case 1:
		doSimpleton() ;			// Simpleton test
		break ; 

	case 2:
		doSmarty() ;			// Smarty test
		break ; 

	case 3: 	
		doBoundedBuffer() ;		// Bounded buffer test
		break ; 

	default:
		cout << "Fatal: Experiment #" << testCase << " is not yet available" << endl ; 
		gracefulExit( 1) ; 
		break ; 
	}

	gracefulExit( 0) ; 
	return 0 ; 
}

///////
// doSimpleton()
// Create the threads for the simpleton test. 
// There is no synchronization between threads here.
//
void doSimpleton() { 
	HANDLE		threadHandles[ MAX_THREADS] ; 
	DWORD		threadIds[ MAX_THREADS] ; 
	int			i ; 

	// Get the number of threads to create
	numThreads = queryUserForInt( "Enter number of threads to create") ; 
	if( numThreads > MAX_THREADS) { 
		cout << "Fatal: Number of thread is limited to " << MAX_THREADS << endl ; 
		gracefulExit( 1) ; 
	}

	// No synchronization objects used

	// Create the n "simpleton" threads
	for( i = 0; i < numThreads; i++) { 
		threadHandles[i] = CreateThread( NULL, 0, simpleton, 
			NULL, 0, &threadIds[i]) ; 
	}

	// Wait for all threads to complete, then return
	WaitForMultipleObjects( numThreads, threadHandles, true, INFINITE) ; 
}

///////
// simpleton()
// The simpleton thread function has no synchronization, so it's output, 
// when multiple copies are running, will probably appear jumbled.
//
DWORD WINAPI simpleton( LPVOID) { 
	int		myCount ; 

	// Assign thread's count and output that we are running
	myCount = threadCount++ ; 
	cout << "Simpleton thread #" << myCount <<  " is running" << endl ; 
	return 0 ; 
}

/////// 
// doSmarty() 
// This is similar to the simpleton test, but we will use a mutex
// to synchronize output, so that it doesn't get all jumbled.
//
void doSmarty() { 
	HANDLE		threadHandles[ MAX_THREADS] ; 
	DWORD		threadIds[ MAX_THREADS] ; 
	int			i ; 

	// Get the number of threads to create
	numThreads = queryUserForInt( "Enter number of threads to create") ; 
	if( numThreads > MAX_THREADS) { 
		cout << "Fatal: Number of thread is limited to " << MAX_THREADS << endl ; 
		gracefulExit( 1) ; 
	}

	// **> Create synch objects: a mutex
	mutex = CreateMutex( NULL, FALSE, NULL) ; 

	// **> Create the n smarty threads
	for( i = 0; i < numThreads; i++) { 
		threadHandles[i] = CreateThread( NULL, 0, smarty, 
			NULL, 0, &threadIds[i]) ; 
	}

	// **> Wait for all threads to complete, then return
	WaitForMultipleObjects( numThreads, threadHandles, true, INFINITE) ; 

}

///////
// smarty()
// Implements a smarter thread function that uses a mutex to protect
// its I/O. Also, after the initial I/O, the smarty thread waits for
// all the other smartys to be created and then writes one more time.
//
DWORD WINAPI smarty( LPVOID) { 
	int		myCount ; 

	// **> wait( mutex)
	WaitForSingleObject( mutex, INFINITE) ; 

	// Assign thread's count and output that we are running
	myCount = threadCount++ ; 
	cout << "Smarty thread #" << myCount <<  " is running" << endl ; 

	// **> signal( mutex)
	ReleaseMutex( mutex) ; 

	// **> Wait for all threads to be running
	while( threadCount < numThreads) { 
		Sleep( 0) ; 
	}

	// **> wait( mutex)
	WaitForSingleObject( mutex, INFINITE) ; 

	// Write again, now that all threads are running
	cout << "Smarty thread #" << myCount << ": all threads accounted for" << endl ; 

	// **> signal( mutex)
	ReleaseMutex( mutex) ; 

	return 0 ; 
}

///////
// doBoundedBuffer()
// This test implements the Bounded Buffer problem as it appears on
// Page 206 in our text. A consumer threads and "n" producer threads are 
// created. A mutex and semaphores are used to synchronize communication
// between these threads.
//
void doBoundedBuffer() {
	HANDLE		threadHandles[ MAX_THREADS] ; 
	DWORD		threadIds[ MAX_THREADS] ; 
	int			i, j ; 

	// Do additional user queries: output file, # consumers, 
	// # producers, # buffers, output file
	numConsumers = 1 ; 
	numProducers = queryUserForInt( "Enter number of producers") ; 
	numThreads = numProducers + numConsumers ; 
	if( numThreads > MAX_THREADS) { 
		cout << "Fatal: Total number of threads is limited to " << MAX_THREADS << endl ; 
		gracefulExit( 1) ; 
	}

	numBuffers = queryUserForInt( "Enter number of buffers") ; 
	if( numBuffers > MAX_BUFFERS) { 
		cout << "Fatal: Number of buffers is limited to " << MAX_BUFFERS << endl ; 
		gracefulExit( 1) ; 
	}

	itemsPerProducer = queryUserForInt( "Enter number of items per producer") ; 

	maxProducerTime = queryUserForInt( "Enter max time (in millisecs) to produce an item") ; 

	// Max consumer wait time must be order magnitude greater than
	// max production time, so timeout is not too early.
	// Also, the min wait time for a timeout is 1 second.
	maxConsumerWait = maxProducerTime * 10 ; 
	if( maxConsumerWait < 1000) { 
		maxConsumerWait = 1000 ;	// consumer wait at least 1 second
	}

	// Open the output file for this run
	outFileName = queryUserForString( "Enter output file") ; 
	outFile.open( outFileName.data(), ios::out) ; 
	if( ! outFile) { 
		cout << "Fatal: Can't open output file " << outFileName << endl ; 
		gracefulExit( 1) ; 
	}

	outFile << "Bounded Buffer Test" << endl ; 
	outFile << "   # Consumers: " << numConsumers << endl ; 
	outFile << "   # Producers: " << numProducers << endl ; 
	outFile << "   # Buffers: " << numBuffers << endl ; 
	outFile << "   # Items/Producer: " << itemsPerProducer << endl ; 
	outFile << "   Max time to produce an item (millisecs): " << maxProducerTime << endl ; 
	outFile << endl ; 

	// **> Create synch objects used in the Bounded Buffer test: 
	// a mutex and 2 semaphores
	mutex = CreateMutex( NULL, FALSE, NULL) ; 
	full = CreateSemaphore( NULL, 0, MAX_BUFFERS, NULL) ; 
	empty = CreateSemaphore( NULL, numBuffers, MAX_BUFFERS, NULL) ; 

	// Seed the random number generator, used later on
	srand( (unsigned) time( NULL ) );

	// **> Create the consumer thread, then the producer threads
	for( i = 0; i < numConsumers; i++) { 
		threadHandles[i] = CreateThread( NULL, 0, consumer, 
			(void *) i, 0, &threadIds[i]) ; 
	}
	for( j = 0; j < numProducers; j++) { 
		threadHandles[j+numConsumers] = CreateThread( NULL, 0, producer, 
			(void *) j, 0, &threadIds[j+numConsumers]) ; 
	}

	// **> Wait until all threads are done, then close the output file and return
	WaitForMultipleObjects( numThreads, threadHandles, true, INFINITE) ; 
	outFile.close() ; 
}

///////
// consumer()
// The consumer thread function for the bounded buffer problem.
// The thread waits for full buffers, then munges them, using a mutex
// to perform the operation "safely". Note all consumption to the 
// local output file.
// NOTE: This is pretty much straight from page 207 of our text
//
DWORD WINAPI consumer( LPVOID param) { 
	int			consumerNum = (int) param ; 
	DWORD		ret ; 
	int			curData ; 

	while( 1) { 
		// **> wait( full), wait until a full buffer is available
		// You also may want to consider a timeout here signifying 
		// the end of the test/problem.
		ret = WaitForSingleObject( full, maxConsumerWait) ; 
		if( ret == WAIT_TIMEOUT) { 
			outFile << "Consumer C" << consumerNum << " has timed out!" << endl ; 
			break ; 
		}

		// **> wait( mutex), changing the buffer requires mutex synch
		WaitForSingleObject( mutex, INFINITE) ; 

		// **> Consume! Remove the item from the buffer and 
		// write to the output file so you can follow what's happening.
		if( ! bufferQ.empty()) { 
			curData = bufferQ.front() ; 
			bufferQ.pop() ; 
			outFile << "   C" << consumerNum << " consumed item #" << curData << endl ; 
		}

		// **> signal( mutex), we're done changing the buffer, release mutex
		ReleaseMutex( mutex) ; 

		// **> signal( empty), there's an empty buffer available now
		ReleaseSemaphore( empty, 1, NULL) ; 
	}

	return 0 ; 
}

///////
// producer()
// The producer thread function for the Bounded Buffer problem.
// The thread waits for empty buffers, then fills them with "data".
// A mutex is used to perform the operation "safely".
// NOTE: This is straight from page 207 of our text
//
DWORD WINAPI producer( LPVOID param) { 
	int			producerNum = (int) param ; 
	int			produceCount ; 
	int			produceTime = 0 ; 
	int			data ; 

	for( produceCount = 0; produceCount < itemsPerProducer; produceCount++) { 

		// **> Simulate randomized time to produce the next item
		// using rand() and Sleep()
		if( maxProducerTime > 0) {
			produceTime = (rand() % maxProducerTime) ; 
			Sleep( (DWORD) produceTime) ; 
		}

		// **> wait( empty), wait for an empty buffer
		WaitForSingleObject( empty, INFINITE) ; 

		// **> wait( mutex), changing the buffer requires mutex synch
		WaitForSingleObject( mutex, INFINITE) ; 

		// **> Produce! Add an item to the buffer and 
		// write to the output file so that you can follow what's happeneing
		data = 100 * producerNum + produceCount ;  
		bufferQ.push( data) ; 
		outFile << "P" << producerNum << " produced item #" << data << 
				" after time=" << produceTime << endl ; 

		// **> signal( mutex), we're done changing the buffer, release mutex
		ReleaseMutex( mutex) ; 

		// **> signal( full), there's a full buffer available!
		ReleaseSemaphore( full, 1, NULL) ; 
	}

	return 0 ;
}

///////
// queryUserForInt()
// Query the user for an integer value and return it
int queryUserForInt( char *prompt = 0) { 
	int		iValue ; 

	if( prompt != 0) { 
		cout << prompt << ": " ; 
	}
	else { 
		cout << "Enter integer value: " ; 
	}
	cin >> iValue ; 
	return iValue ; 
}

///////
// queryUserForString()
// Query the user for a string value and return it
//
string queryUserForString( char *prompt = 0) { 
	string		strValue ; 

	if( prompt != 0) { 
		cout << prompt << ": " ; 
	}
	else { 
		cout << "Enter string value: " ; 
	}
	cin >> strValue ; 
	return strValue ; 
}

///////
// gracefulExit()
// Pause, then call exit() to terminate program execution
//
void gracefulExit( int exitCode) { 
	char	ch ; 

	cout << "Hit <enter> to exit" << endl ; 
	cin.get( ch) ; 
	exit( exitCode) ; 
}
